@@ -3,17 +3,14 @@ print greet('Ingy', 42);
 use Inline C => <<'END_OF_C_CODE';
 
 void greet(char* name, int number) {
-  dXSARGS;
+  Inline_Stack_Vars;
   int i;
 
-  sp = mark;
-
+  Inline_Stack_Reset;
   for (i = 0; i < number; i++)
-    XPUSHs(sv_2mortal(newSVpvf("Hello %s!\n", name))); 
+    Inline_Stack_Push(sv_2mortal(newSVpvf("Hello %s!\n", name))); 
 
-  PUTBACK;
+  Inline_Stack_Done;
 }
 
 END_OF_C_CODE
-
-__END__
